#!/usr/bin/env python2.7

# Creates the mendelian CNVs bed and pickle.

import sys
import argparse
import pickle

import numpy as np

import bed
import validation
import parsers
import cnv_struct

def main(args):
    if (args.format == "cnver" and
        args.doc_over == float("-infinity") and
        args.doc_under == float("+infinity")):
        # Default values for cnver
        doc_lower_bound = 0.25
        doc_upper_bound = 3.0

    else:
        doc_lower_bound = args.doc_over
        doc_upper_bound = args.doc_under

    parents_ro = args.parents_ro
    twins_ro = args.twins_ro

    p = validation.get_parser_if_valid(args)
    gains = 0
    losses = 0

    # Get the CNVs
    cnvs = p.get_cnvs()
    inherited, counts = get_inherited(cnvs, twins_ro, parents_ro, 
        doc_lower_bound, doc_upper_bound)

    i = float(counts["good_twin_ro"]) / float(counts["total"])
    print ("The proportion of twin-overlapping CNVs is {}".format(i))

    i = float(len(inherited)) / counts["good_twin_ro"]
    print ("The proportion of twin-overlapping "
           "CNVs that are inherited is: {}").format(i)

    i = float(len(inherited)) / counts["total"]
    print "The proportion of inherited CNVs is: {} ({}/{})".format(i, 
        len(inherited), counts["total"])
    # Write pickle
    if args.format:
        fn = "inherited_{}_{}.pickle".format(p.family_id, args.format)
    else:
        fn = "inherited_{}.pickle".format(p.family_id)

    with open(fn, "w") as f:
        pickle.dump(inherited, f)

def get_inherited(cnvs, 
                  twins_ro, 
                  parents_ro,
                  doc_lower_bound = float("-infinity"),
                  doc_upper_bound = float("+infinity")):
    inherited = []
    counts = {"gain": 0, 
              "loss": 0, 
              "good_twin_ro": 0,
              "total": 0}

    other = "twin2"
    for sample in cnvs:
        if sample == "twin1":
            for chr in cnvs[sample]:
                for cnv in cnvs[sample][chr]:
                    counts["total"] += 1
                    twin_ro, twin_cnv = cnv_struct.ro_to_best_in_list(
                                                                  cnv,
                                                                  cnvs[other][chr],
                                                                  return_cnv = True
                                                               )
                    twin_ro = min(twin_ro)

                    father_ro = cnv_struct.ro_to_best_in_list(
                                                                cnv,
                                                                cnvs["father"][chr]
                                                             )

                    mother_ro = cnv_struct.ro_to_best_in_list(
                                                                cnv,
                                                                cnvs["mother"][chr]
                                                             )

                    # This should be above the ro threshold.
                    max_min_parent_ro = max(min(mother_ro),
                                            min(father_ro))

                    add = True
                    if (cnv.doc and
                        cnv.doc < doc_lower_bound or 
                        cnv.doc > doc_upper_bound):
                        add = False
                    elif twin_ro < twins_ro:
                        add = False
                    elif max_min_parent_ro < parents_ro:
                        add = False

                    # if (min(mother_ro) > parents_ro and 
                    #     min(father_ro) > parents_ro):
                    #     # ---- ou ++++
                    #     add = False
                    #     counts["total"] -= 1

                    # Statistics.
                    if twin_ro > twins_ro:
                        counts["good_twin_ro"] += 1

                    if add:
                        # Create a new CNV representing this inherited CNV.
                        assert cnv.type == twin_cnv.type
                        assert cnv.chr == twin_cnv.chr

                        pos = "chr{}:{}-{}".format(
                                                    chr,
                                                    min(twin_cnv.start, cnv.start),
                                                    max(twin_cnv.end, cnv.end),
                                                  )

                        new_cnv = cnv_struct.cnv(
                                                    pos = pos,
                                                    type = cnv.type,
                                                    algo = cnv.algo,
                                                    source = None,
                                                    confidence = getattr(cnv, 
                                                        "confidence", None)
                                                )
                        inherited.append(new_cnv)
                        if new_cnv.type in ("gain", "loss"):
                            counts[new_cnv.type] += 1

    return inherited, counts


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Computes the average and std CNV size.")

    parser = validation.add_pickle_args(parser)

    parser.add_argument("--format",
                        type=str,
                        choices=parsers.__all__,
                        help="The file format for the parser.")

    parser.add_argument("--base_dir",
                        type=str,
                        help=("The base directory for the calls. A directory with "
                              "four subfolders starting with father, mother, twin1 "
                              "and twin2 are expected."))

    parser.add_argument("--doc_under",
                        type=float,
                        help="Maximum doc value (for filtering)",
                        default=float("+infinity"),
                       )

    parser.add_argument("--doc_over",
                        type=float,
                        help="Minimum doc value (for filtering)",
                        default=float("-infinity"),
                       )

    parser.add_argument("--parents_ro",
                        type=float,
                        help=("Only consider CNVs that overlap with this "
                              "minimum threshold between the twins and "
                              "only one parent (to reduce false positives) "
                              "default %(default).1f."),
                        default=0.5
                       )

    parser.add_argument("--twins_ro",
                        type=float,
                        help=("Only consider CNVs that overlap with this "
                              "minimum threshold between the two twins "
                              "default %(default).1f."),
                        default=0.9
                       )

    main(parser.parse_args()) 
