#!/usr/bin/env python2.7

import argparse

import matplotlib
#matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy as np

import generate_cnv_graph as g

def main(args):
    pileups = (args.twin1_pileup,
               args.twin2_pileup,
               args.mother_pileup,
               args.father_pileup)

    pileups = tuple([open(f) for f in pileups])

    plot_region(args.chr, args.start, args.end, pileups, args.hstart, args.hend)

def plot_region(chromo, start, end, pileups, hstart = None, hend = None):
    g.normalize_pileups_starting_position(*pileups)
    indexes = g.create_seek_index(pileups)

    eof = g.seek_to(start - 1, chromo, pileups, indexes)
    lines = [i.readline().split("\t") for i in pileups]

    coverages = {
        "twin1": [],    
        "twin2": [],
        "mother": [],
        "father": []
    }

    while int(lines[0][1]) <= end:
        # Assuming that pileups is twin1, twin2, mother, father
        coverages["twin1"].append(int(lines[0][3]))
        coverages["twin2"].append(int(lines[1][3]))
        coverages["mother"].append(int(lines[2][3]))
        coverages["father"].append(int(lines[3][3]))

        lines = [i.readline().split("\t") for i in pileups]

    for i in coverages:
        coverages[i] = np.array(coverages[i])

    # Compute average in and out of region if hstart and hend are given.
    average = {
        "twin1": [], "twin2": [], "mother": [], "father": []
    }

    if hstart and hend:
        for i in coverages:
            pre = np.mean(coverages[i][0:hstart - start])
            region = np.mean(coverages[i][hstart - start : hend - start])
            post = np.mean(coverages[i][hend - start :])
            
            average[i] = np.array([pre, region, post])

    # Plotting
    fig, axes = plt.subplots(2, 2, figsize=(11, 8.5), sharey = True, sharex = True)
    fig.suptitle("Chromosome {}".format(chromo), weight="bold")
    fig.subplots_adjust(left=0.07, bottom=0.07, right=0.96, top=0.94,
                        wspace=0.19, hspace=0.09)
    named_axes = {
        "twin1": axes[0][0],
        "twin2": axes[0][1],
        "mother": axes[1][0],
        "father": axes[1][1],
    }

    for i in named_axes.keys():
        named_axes[i].plot(
                               np.arange(start, end + 100)[:len(coverages[i])],
                               coverages[i],
                               "k-",
                           )
        if i in ("mother", "father"):
            named_axes[i].set_xlabel("Position (Mb)")
        named_axes[i].set_ylabel("Coverage in {}".format(i))
        named_axes[i].set_xticks(np.array([start, end]))
        named_axes[i].set_xticklabels(np.round(np.array([start, end]) / 1e6, 3))

        for tick in (named_axes[i].yaxis.get_major_ticks() + 
                     named_axes[i].xaxis.get_major_ticks()):
            tick.label.set_fontsize(10)

        if hstart and hend:
            # Highlight region
            named_axes[i].axvspan(hstart, hend, facecolor="#cccccc", alpha=0.5)

            # Show the average doc
            # x-pos entre 0 et 1
            hstart_norm = (hstart - start) / float(end - start)
            hend_norm = (hend - start) / float(end - start)

            named_axes[i].axhline(average[i][0], 0, hstart_norm, color="#FF4444")
            named_axes[i].axhline(average[i][1], hstart_norm, hend_norm, color="#FF4444")
            named_axes[i].axhline(average[i][2], hend_norm, 1, color="#FF4444")

    plt.savefig("region_plot.pdf")

if __name__ == "__main__":
    PARSER = argparse.ArgumentParser("Plots a region for a family given pileup "
                                   "files.")

    for i in ("twin1", "twin2", "mother", "father"):
        PARSER.add_argument(
            "--{}_pileup".format(i),
            type = str,
            required = True,
            help = "Path to pileup for {}".format(i)             
        )

    PARSER.add_argument("--start",
                        type=int,
                        help="Start position.",
                        required = True,
                       )

    PARSER.add_argument("--chr",
                        type=int,
                        help="Chromosome",
                        required = True,
                       )

    PARSER.add_argument("--hend",
                        type=int,
                        help="Use to highlight a region (e.g.) a CNV.",
                        default=None
                       )

    PARSER.add_argument("--hstart",
                        type=int,
                        help="Use to highlight a region (e.g.) a CNV.",
                        default=None
                       )

    PARSER.add_argument("--end",
                        type=int,
                        help="End position.",
                        required = True,
                       )

    main(PARSER.parse_args())
